# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
About dialog.
"""
from epoptes import __version__
from epoptes.ui.common import locate_resource
from gi.repository import Gtk


class About:
    """Show an about dialog."""
    def __init__(self, parent):
        builder = Gtk.Builder()
        builder.add_from_file(locate_resource('about.ui'))
        builder.connect_signals(self)
        self.dialog = builder.get_object('dlg_about')
        self.dialog.set_transient_for(parent)
        self.dialog.set_version(__version__)

    def run(self):
        """Show the dialog, then hide it so that it may be reused."""
        self.dialog.run()
        self.dialog.hide()



# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Execute command dialog.
"""
from epoptes.common import config
from epoptes.ui.common import locate_resource
from gi.repository import Gtk


class ExecCommand:
    """Load the dialog and settings into local variables."""
    def __init__(self, parent):
        builder = Gtk.Builder()
        builder.add_from_file(locate_resource('exec_command.ui'))
        self.dialog = builder.get_object('dlg_exec_command')
        self.dialog.set_transient_for(parent)
        self.cbt_command = builder.get_object('cbt_command')
        self.ent_command = self.cbt_command.get_child()
        self.btn_execute = builder.get_object('btn_execute')
        builder.connect_signals(self)
        for cmd in config.history:
            self.cbt_command.append_text(cmd)

    def run(self):
        """Show the dialog, then hide it so that it may be reused.
        Return the command.
        """
        reply = self.dialog.run()
        if reply == 1:
            result = self.ent_command.get_text().strip()
            if result in config.history:
                config.history.remove(result)
            config.history.insert(0, result)
            config.write_plain_file(
                config.expand_filename('history'), config.history)
        else:
            result = ''
        self.dialog.hide()

        return result

    def on_ent_command_changed(self, ent_command):
        """Enable execute only when the command is not empty."""
        self.btn_execute.set_sensitive(ent_command.get_text().strip())



# This file is part of Epoptes, https://epoptes.org
# Copyright 2015-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Notifications.
Keep at most 10 lines of headings and messages.
A new bubble is created after 10 secs of inactivity.
That means that the oldest event may show for up to 9*10 seconds or so.

Example bubble layout:
[icon] Epoptes
*Connected:*
  user1 on host1
  user2 on host2
*Disconnected:*
  user3 from host3
*Shut down:*
  host4
*Connected:*
  user5 on host5

Notes for current server implementations (get_server_info, get_server_caps):
* bionic-ubuntu: no multiple bubbles,
    shows up to only 6 lines when hovered, merges '\n' otherwise,
    keeps history in the clock (one entry for all updates)
  (True, ret_name='gnome-shell', ret_vendor='GNOME',
    ret_version='3.28.1', ret_spec_version='1.2')
  ['actions', 'body', 'body-markup', 'icon-static', 'persistence', 'sound']
* bionic-mate: supports multiple bubbles
  (True, ret_name='Notification Daemon', ret_vendor='MATE',
    ret_version='1.20.0', ret_spec_version='1.1')
  ['actions', 'action-icons', 'body', 'body-hyperlinks', 'body-markup',
    'icon-static', 'sound']
* bionic-kubuntu: shows scrollbar after 10 messages,
    keeps history in the notification area (one entry per update!)
  (True, ret_name='Plasma', ret_vendor='KDE',
    ret_version='2.0', ret_spec_version='1.1')
  ['actions', 'body', 'body-hyperlinks', 'body-markup', 'icon-static']
* bionic-lubuntu (xfce4-notifyd): same as bionic-xubuntu
* bionic-xubuntu (xfce4-notifyd): closes on click (no [x] button)
  (True, ret_name='Xfce Notify Daemon', ret_vendor='Xfce',
    ret_version='0.4.2', ret_spec_version='1.2')
  ['actions', 'body', 'body-hyperlinks', 'body-markup', 'icon-static',
    'x-canonical-private-icon-only']
* stretch-gnome: same as bionic-ubuntu
* stretch-gnome (with notification-daemon -r): no multiple bubbles
  (True, ret_name='Notification Daemon', ret_vendor='GNOME',
    ret_version='3.20.0', ret_spec_version='1.2')
  ['actions', 'action-icons', 'body', 'body-hyperlinks', 'body-markup',
    'icon-static', 'persistence', 'sound']
* sylvia-cinnamon:
  (True, ret_name='cinnamon', ret_vendor='GNOME',
    ret_version='3.6.7', ret_spec_version='1.2')
  ['actions', 'action-icons', 'body', 'body-markup', 'icon-static',
    'persistence']
* xenial-ubuntu:
  (True, ret_name='notify-osd', ret_vendor='Canonical Ltd',
    ret_version='1.0', ret_spec_version='1.1')
  ['body', 'body-markup', 'icon-static', 'image/svg+xml',
    'x-canonical-private-synchronous', 'x-canonical-append',
    'x-canonical-private-icon-only', 'x-canonical-truncation',
    'private-synrchronous', 'append', 'private-icon-only', 'truncation']
"""
import time

from epoptes.core import logger
from epoptes.ui.common import gettext as _
from gi.repository import GLib, Notify

LOG = logger.Logger(__file__)

class NotifyQueue:
    """A special queue that keeps at most 10 lines of headings and messages."""
    initialized = None

    def __init__(self, summary, icon):
        """Initialize Notify and local variables."""
        if not NotifyQueue.initialized:
            NotifyQueue.initialized = True
            if not Notify.init("Epoptes"):
                raise ImportError(_('Could not initialize notifications!'))
        self.summary = summary
        self.icon = icon
        self.items = []
        # The heading of the last item enqueued
        self.last_heading = ''
        self.last_time = time.time()
        self.notification = None
        self.shown_error = False

    def enqueue(self, heading, msg):
        """Add a new message to the queue, and show it."""
        # Create a new bubble if 10 secs have passed from the last notification
        now = time.time()
        if now - self.last_time > 10:
            self.__init__(self.summary, self.icon)
        self.last_time = now
        if heading != self.last_heading:
            self.last_heading = heading
            self.items.append("<b>{}</b>".format(heading))
        self.items.append("  {}".format(msg))
        # One enqueue may insert both heading and msg, needing 2 dequeues
        while len(self.items) > 10:
            self.dequeue()
        self.show()

    def dequeue(self):
        """Delete the first message after the first heading.
        If the section is now empty, delete the first heading as well."""
        del self.items[1]
        if (not self.items) \
                or (len(self.items) > 1 and self.items[1].startswith('<b>')):
            del self.items[0]

    def to_string(self):
        """Return a string of all queued messages."""
        return '\n'.join(self.items)

    @staticmethod
    def new_notification(summary, body, icon):
        """Create a new notification object. Called by show()."""
        result = Notify.Notification.new(summary, body, icon)
        result.set_urgency(Notify.Urgency.LOW)
        result.set_hint_string('desktop-entry', 'epoptes')
        # This tells GNOME not to log the notification
        result.set_hint_string('transient', 'true')
        return result

    def show(self):
        """Create, update, or re-create the notification, and show it."""
        if self.notification is None:
            self.notification = self.new_notification(
                self.summary, self.to_string(), self.icon)
        elif not self.notification.update(
                self.summary, self.to_string(), self.icon):
            self.notification = self.new_notification(
                self.summary, self.to_string(), self.icon)
        try:
            self.notification.show()
        # pylint: disable=catching-non-exception
        except GLib.Error as exc:
            # If a notification daemon isn't running, show an error once
            if not self.shown_error:
                self.shown_error = True
                LOG.e("Error while showing notification, is a daemon running?",
                      "\n", str(exc))


def main():
    """Run a notifications test from the command line.
    If epoptes isn't installed, `import epoptes.ui.common` must be removed."""
    print("get_server_info =", Notify.get_server_info())
    print("get_server_caps =", Notify.get_server_caps())
    events = [
        ("Connected:", "user1 on host1"),
        ("Connected:", "user2 on host2"),
        ("Disconnected:", "user3 from host3"),
        ("Shut down:", "host4"),
        ("Connected:", "user5 on host5"),
        ("Connected:", "user6 on host6"),
        ("Shut down:", "host7"),
        ("Shut down:", "host8"),
        ("Connected:", "user9 on host9"),
        ("Connected:", "user10 on host10")
    ]
    i = 0
    notq = NotifyQueue('Epoptes', 'dialog-information')
    while i < len(events):
        notq.enqueue(events[i][0], events[i][1])
        time.sleep(1)
        i += 1
        if i == 5:
            # Wait for the bubble to autoclose, then a new one should show up
            time.sleep(10)


if __name__ == '__main__':
    main()



# This file is part of Epoptes, https://epoptes.org
# Copyright 2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Define required gi package versions in a common place, and install gettext.

Rationale:
gi requires something like:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
This conflicts with https://www.python.org/dev/peps/pep-0008/#imports
and triggers pycodestyle's "E402 module level import not at top of file".
The following is a bit better:
    import sys  # Import standard library modules

    import twisted  # Import third party modules

    from epoptes.ui.common import gettext as _  # Import local modules
    from gi.repository import Gtk, Gdk
That last line "only" triggers pylint's "wrong-import-position" once.
"""
import errno
import gettext
import locale
import os

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
gi.require_version('Notify', '0.7')
gettext.textdomain('epoptes')
locale.textdomain('epoptes')
gettext = gettext.gettext


def locate_resource(filename, absolute=True):
    """Search for filename in some known paths."""
    # Use recursion for absolute instead of multiple ifs:
    if absolute:
        return os.path.abspath(locate_resource(filename, False))
    test = filename
    if os.path.isfile(test):
        return test
    test = "/usr/share/epoptes/" + os.path.basename(filename)
    if os.path.isfile(test):
        return test
    test = "/usr/share/epoptes/images/" + os.path.basename(filename)
    if os.path.isfile(test):
        return test
    raise FileNotFoundError(
        errno.ENOENT, os.strerror(errno.ENOENT), filename)



# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Protocol for epoptesd.py <=> uiconnection.py communication.
"""
import os
import sys
import tempfile

from twisted.internet import protocol
from twisted.protocols import amp

from epoptes.core import logger
from epoptes.daemon import commands, exchange

# stdout in epoptesd/bashplex/guiplex is handled by twisted and goes to syslog.
# stderr isn't, so tell Logger to use stdout.
LOG = logger.Logger(__file__, sys.stdout)


class GUI(amp.AMP):
    """Protocol for epoptesd.py <=> uiconnection.py communication."""

    def connectionMade(self):
        """Override BaseProtocol.connectionMade."""
        super().connectionMade()
        exchange.known_guis.append(self)

    def connectionLost(self, reason):
        """Override AMP.connectionLost."""
        super().connectionLost(reason)
        exchange.known_guis.remove(self)

    def client_connected(self, handle):
        """Remotely call uiconnection.py->Daemon.client_connected."""
        self.callRemote(commands.ClientConnected, handle=handle)

    def client_disconnected(self, handle):
        """Remotely call uiconnection.py->Daemon.client_connected."""
        self.callRemote(commands.ClientDisconnected, handle=handle)

    @commands.EnumerateClients.responder
    def enumerate_clients(self):
        """Remotely called from uiconnection.py->Daemon.enumerate_clients."""
        return {'handles': sorted(exchange.known_clients.keys())}

    @commands.ClientCommand.responder
    def client_command(self, handle, command):
        """Remotely called from uiconnection.py->Daemon.command."""
        if handle not in exchange.known_clients:
            LOG.e("Unknown client %s, can't execute %s" % (handle, command))
            # raise ValueError("Unknown client")
            return {'filename': '', 'result': b''}

        dfr = exchange.known_clients[handle].command(command)

        def send_result(result):
            """Callback for bashplex.py->DelimitedBashReceiver.command."""
            if len(result) < 65000:
                return {'filename': '', 'result': result}
            tmpf = tempfile.NamedTemporaryFile('wb', dir="/run/epoptes",
                                               delete=False)
            tmpf.write(result)
            os.fchmod(tmpf.file.fileno(), 0o660)
            return {'filename': tmpf.name, 'result': b''}

        dfr.addCallback(send_result)
        return dfr


class GUIFactory(protocol.ServerFactory):
    """The factory for the GUI protocol."""
    protocol = GUI



#!/usr/bin/python3
# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Create a thumbshot of the current screen.
"""
import os
import sys

import cairo

from _common import gettext as _
from gi.repository import Gdk


def thumbshot(width, height):
    """Return a thumbshot of the current screen as bytes."""
    root = Gdk.get_default_root_window()
    if root is None:
        raise RuntimeError('Cannot find the root window, is xorg running?')
    geometry = root.get_geometry()
    surface = cairo.ImageSurface(cairo.FORMAT_RGB24, width, height)
    ctx = cairo.Context(surface)
    # TODO: check if this actually does client-size resizing
    ctx.scale(float(width) / geometry.width, float(height) / geometry.height)
    Gdk.cairo_set_source_window(ctx, root, 0, 0)
    ctx.paint()

    # TODO: is a pixbuf necessary, or can we get the bytes from the surface?
    pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, width, height)
    rowst = pixbuf.get_rowstride()
    pixels = pixbuf.get_pixels()

    return (b"%i\n%ix%i\n" % (rowst, width, height)
            + pixels
            # TODO: the last padding isn't included, so do it manually
            + b"\0"*(rowst*height - len(pixels)))


def main():
    """Run the module from the command line."""
    if len(sys.argv) == 3:
        sys.stdout.buffer.flush()
        sys.stdout.buffer.write(thumbshot(int(sys.argv[1]), int(sys.argv[2])))
        sys.stdout.buffer.flush()
    else:
        print(_("Usage: {} width height").format(
            os.path.basename(__file__)), file=sys.stderr)
        exit(1)


if __name__ == '__main__':
    main()



# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Variables and functions to keep track of known clients and GUIs.
"""
import sys
from epoptes.core import logger

# stdout in epoptesd/bashplex/guiplex is handled by twisted and goes to syslog.
# stderr isn't, so tell Logger to use stdout.
LOG = logger.Logger(__file__, sys.stdout)

known_clients = {}
timed_out_clients = {}  # To distinguish reconnections from new connections
known_guis = []


def client_connected(handle, client):
    """Called from bashplex.py->connectionMade."""
    known_clients[handle] = client
    # Notify all known GUIs that an epoptes-client was connected
    for gui in known_guis:
        gui.client_connected(handle)


def client_disconnected(handle):
    """Called from bashplex.py->connectionLost."""
    if handle not in known_clients:
        LOG.e("Disconnect from unknown client: %s" % handle)
        return
    del known_clients[handle]
    # Notify all known GUIs that an epoptes-client was disconnected
    for gui in known_guis:
        gui.client_disconnected(handle)


def client_timed_out(handle):
    """Called from bashplex.py->ping_timed_out."""
    timed_out_clients[handle] = known_clients[handle]
    client_disconnected(handle)


def client_reconnected(handle):
    """Called from bashplex.py->ping_response."""
    client_connected(handle, timed_out_clients[handle])
    del timed_out_clients[handle]



# This file is part of Epoptes, https://epoptes.org
# Copyright 2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Similar to logging.Logger, but simpler.
"""
# TODO: implement debug levels and environment interface.
import os
import sys


class Logger:
    """Similar to logging.Logger, but simpler."""
    def __init__(self, name=None, file=sys.stderr):
        """Check if debugging is activated for "name"."""
        debug = os.getenv('DEBUG', '')
        if debug and name.endswith(debug):
            func = self.stderr
        else:
            func = self.null
        # debug, info, warning, error, critical
        self.d = func
        self.i = func
        self.w = self.stderr
        self.e = self.stderr
        self.c = self.stderr
        self.file = file

    def null(self, *_args, end='\n'):
        """The self.[cdeiw] variables point here when debugging is disabled."""
        pass

    def stderr(self, *args, end='\n'):
        """The self.[cdeiw] variables point here when debugging is enabled."""
        if self:
            print(*args, end=end, file=self.file)



# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
GUI-side part of the gui.py <=UNIX=> guiplex.py protocol.
"""
import os

from twisted.protocols import amp

from epoptes.daemon import commands


class Daemon(amp.AMP):
    """GUI-side part of the gui.py <=UNIX=> guiplex.py protocol."""
    def __init__(self, client):
        super().__init__()
        self.client = client

    def connectionMade(self):
        """Override BaseProtocol.connectionMade."""
        super().connectionMade()
        # We've connected the GUI to the Daemon.
        # Call gui.py->EpoptesGui.connected to store a reference to daemon.
        self.client.connected(self)

    def connectionLost(self, reason):
        """Override AMP.connectionLost."""
        super().connectionLost(reason)
        self.client.disconnected(self)

    @commands.ClientConnected.responder
    def client_connected(self, handle):
        """Remotely called from guiplex.py->GUI.client_connected."""
        # Call gui.py->EpoptesGui.amp_client_connected.
        self.client.amp_client_connected(handle)
        return {}

    @commands.ClientDisconnected.responder
    def client_disconnected(self, handle):
        """Remotely called from guiplex.py->GUI.client_disconnected."""
        # Call gui.py->EpoptesGui.amp_client_disconnected.
        self.client.amp_client_disconnected(handle)
        return {}

    def enumerate_clients(self):
        """Remotely call guiplex.py->GUI.enumerate_clients."""
        dfr = self.callRemote(commands.EnumerateClients)
        dfr.addCallback(lambda r: r['handles'])
        return dfr

    def command(self, handle, command):
        """Remotely call guiplex.py->GUI.command."""
        dfr = self.callRemote(commands.ClientCommand,
                              handle=handle,
                              command=command)

        def got_result(response):
            """Callback for guiplex.py->GUI.command."""
            filename = response['filename']
            if filename:
                result = open(filename, 'rb').read()
                os.unlink(filename)
                return result
            return response['result']

        dfr.addCallback(got_result)
        return dfr



# This file is part of Epoptes, https://epoptes.org
# Copyright 2016-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Network benchmark.
"""
# TODO: either switch to iperf3, or reimplement it with python/twisted.
# iperf2 doesn't work behind NAT and has several issues, for example:
# https://sourceforge.net/p/iperf2/discussion/general/thread/db0fed22/
from epoptes.common.constants import C_INSTANCE
from epoptes.core import spawn_process
from epoptes.ui.common import gettext as _, locate_resource
from gi.repository import GLib, Gtk


def humanize(value, decimal=1, unit=''):
    """Convert bits to [KMGTPEZ]bits."""
    value = float(value)
    for prefix in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(value) < 1000:
            value = round(value, decimal)
            return "%.*f %s%s" % (decimal, value, prefix, unit)
        value = value / 1000
    return "%.*f %s%s" % (decimal, value, 'Y', unit)


class Benchmark:
    """Network benchmark."""
    def __init__(self, parent, execute):
        self.clients = {}
        self.countdown_event = None
        self.execute = execute
        self.iperf = None
        self.parent = parent
        self.results = {}
        self.spawn_process = spawn_process.SpawnProcess(self.on_iperf_exit)
        self.timeleft = 0
        builder = Gtk.Builder()
        builder.add_from_file(locate_resource('benchmark.ui'))
        builder.connect_signals(self)
        self.dlg_message = builder.get_object('dlg_message')
        self.dlg_benchmark = builder.get_object('dlg_benchmark')
        self.adj_seconds = builder.get_object('adj_seconds')
        self.box_seconds = builder.get_object('box_seconds')
        self.spb_seconds = builder.get_object('spb_seconds')
        self.box_countdown = builder.get_object('box_countdown')
        self.lbl_countdown = builder.get_object('lbl_countdown')
        self.btn_start = builder.get_object('btn_start')
        self.btn_stop = builder.get_object('btn_stop')
        self.dlg_results = builder.get_object('dlg_results')
        self.lss_results = builder.get_object('lss_results')
        self.tvc_upload = builder.get_object('tvc_upload')
        self.tvc_download = builder.get_object('tvc_download')
        self.crt_upload = builder.get_object('crt_upload')
        self.crt_download = builder.get_object('crt_download')
        self.lbl_avg_down = builder.get_object('lbl_avg_down')
        self.lbl_avg_up = builder.get_object('lbl_avg_up')
        self.lbl_avg_down = builder.get_object('lbl_avg_down')
        self.lbl_avg_up = builder.get_object('lbl_avg_up')
        self.lbl_total_down = builder.get_object('lbl_total_down')
        self.lbl_total_up = builder.get_object('lbl_total_up')
        self.box_partial_results = builder.get_object('box_partial_results')
        self.dlg_message.set_transient_for(self.parent)
        self.dlg_benchmark.set_transient_for(self.parent)
        self.dlg_results.set_transient_for(self.parent)
        self.tvc_upload.set_cell_data_func(self.crt_upload, self.data_func, 1)
        self.tvc_download.set_cell_data_func(
            self.crt_download, self.data_func, 2)

    def warning_message(self, msg):
        """Show a warning dialog."""
        self.dlg_message.set_property("message-type", Gtk.MessageType.WARNING)
        self.dlg_message.set_title(_("Warning"))
        self.dlg_message.set_markup(msg)
        self.dlg_message.run()

    def error_message(self, msg):
        """Show an error dialog."""
        self.dlg_message.set_property("message-type", Gtk.MessageType.ERROR)
        self.dlg_message.set_title(_("Error"))
        self.dlg_message.set_markup(msg)
        self.dlg_message.run()

    def on_dlg_message_close(self, _widget, _event=None):
        """Handle btn_close_message.clicked and dlg_message.delete_event."""
        self.dlg_message.hide()

    def run(self, clients):
        """Show the dialog, then hide it so that it may be reused."""
        self.clients = {}
        # This can happen on an empty group.
        # Btw, "if not clients" is wrong as it's a Gtk.ListStore, not a dict.
        # pylint: disable=len-as-condition
        if len(clients) == 0:
            self.warning_message(
                _('There are no selected clients to run the benchmark on.'))
            return

        # Check if offline clients or clients with no root client are selected
        off = []
        for client in clients:
            inst = client[C_INSTANCE]
            if inst.hsystem:
                self.clients[inst.hsystem.split(':')[0]] =\
                    (inst.hsystem, inst.get_name())
            else:
                off.append(inst.get_name())

        # Now self.clients is the list of clients that can run the benchmark
        if not self.clients:
            self.warning_message(
                _('All of the selected clients are either offline,'
                  ' or do not have epoptes-client running as root.'))
            return
        if off:
            self.warning_message(
                _('The following clients will be excluded from the benchmark'
                  ' because they are either offline, or do not have'
                  ' epoptes-client running as root.')
                + '\n\n' + ', '.join(off))

        self.box_seconds.set_visible(True)
        self.box_countdown.set_visible(False)
        self.btn_start.set_visible(True)
        self.btn_stop.set_visible(False)
        self.dlg_benchmark.run()

    def on_btn_start_clicked(self, _widget):
        """Handle btn_start.clicked event."""
        seconds = int(self.adj_seconds.get_value())
        self.spawn_process.spawn('iperf -s -xS -yC'.split(),
                                 timeout=(seconds + 3),
                                 lines_max=2*len(self.clients))
        for client in self.clients:
            handle = self.clients[client][0]
            # Half time for upload speed and half for download
            self.execute(handle, 'start_benchmark %d' % int(seconds/2))
        self.timeleft = seconds
        self.box_seconds.set_visible(False)
        self.box_countdown.set_visible(True)
        self.btn_start.set_visible(False)
        self.btn_stop.set_visible(True)
        self.lbl_countdown.set_text(_("Benchmark finishing in %d seconds...")
                                    % self.timeleft)
        self.countdown_event = GLib.timeout_add(1000, self.update_countdown)

    def update_countdown(self):
        """Update the countdown label."""
        self.timeleft -= 1
        if self.timeleft >= 0:
            self.lbl_countdown.set_text(
                _("Benchmark finishing in %d seconds...") % self.timeleft)
        else:
            self.lbl_countdown.set_text(
                _("Some clients didn't respond in time!") + "\n"
                + _("Waiting for %d more seconds...") % (self.timeleft + 3))

        # Always recall; the timeout will be cancelled in on_iperf_exit.
        return True

    def parse_iperf_output(self, out_data):
        """Parse 'output' as a string of single or multiple lines of CSV in the
        form of [timestamp, server_ip, port, client_ip, port, id, from-to
        transfered(Bytes), bandwidth(bps)] and populate a dict of client_ip:
         [upload Mbps, download Mbps] pairs storing it in self.results.
        """
        self.results = {}
        data = out_data.strip().split()
        for line in data:
            values = line.split(',')
            if len(values) != 9:
                continue
            client_ip = values[3]
            client_port = values[4]  # will be 5001 if the client is receiving
            bandwidth = int(values[8])
            if client_ip in self.clients:
                if client_ip not in self.results:
                    self.results[client_ip] = [0, 0]

                if client_port == "5001":
                    # Download (bits/s)
                    self.results[client_ip][1] = int(bandwidth)
                else:
                    # Upload (bits/s)
                    self.results[client_ip][0] = int(bandwidth)

    @staticmethod
    def data_func(_column, cell, model, itr, index):
        """Convert model's glong to text, to display humanized units."""
        bps = model[itr][index]
        if bps <= 0:
            cell.set_property("text", "—")
        else:
            cell.set_property("text", humanize(bps, unit='bps'))

    def on_iperf_exit(self, out_data, err_data, reason):
        """The benchmark has finished, show the results dialog."""
        GLib.source_remove(self.countdown_event)
        self.box_seconds.set_visible(True)
        self.box_countdown.set_visible(False)
        self.btn_start.set_visible(True)
        self.btn_stop.set_visible(False)

        if reason == "stopped":
            self.btn_stop.set_sensitive(True)
            return
        elif reason == "closed":
            return

        self.dlg_benchmark.hide()
        self.parse_iperf_output(out_data.decode("utf-8"))
        if not self.results:
            msg = _("Did not get measurements from any of the clients."
                    " Check your network settings.")
            if err_data:
                msg += "\n\n" + err_data.decode("utf-8")
            self.error_message(msg)
            return

        # At this point we do have some results, so show dlg_results
        total_up = 0
        total_down = 0
        self.lss_results.clear()
        # List all the clients regardless of if we received measurements
        for client_ip in self.clients:
            client_name = self.clients[client_ip][1]
            if client_ip in self.results:
                upload, download = self.results[client_ip]
            else:
                upload, download = (0, 0)
            self.lss_results.append([client_name, upload, download])
            total_up += upload
            total_down += download

        clients_n = len(self.clients)
        self.lbl_avg_up.set_text(
            humanize(total_up / clients_n, unit='bps'))
        self.lbl_total_up.set_text(humanize(total_up, unit='bps'))
        self.lbl_avg_down.set_text(
            humanize(total_down / clients_n, unit='bps'))
        self.lbl_total_down.set_text(humanize(total_down, unit='bps'))

        self.box_partial_results.set_visible(
            self.spawn_process.lines_count != 2*len(self.clients))
        self.dlg_results.run()

    def on_btn_stop_clicked(self, _widget):
        """Handle btn_stop.clicked event."""
        self.btn_stop.set_sensitive(False)
        self.spawn_process.stop('stopped')

    def on_dlg_benchmark_close(self, _widget, _event=None):
        """Handle btn_close_benchmark.clicked, dlg_benchmark.delete_event."""
        # For simplicity, we assume that the user isn't fast enough
        # to close the dialog and reopen it before on_exit is called.
        if self.spawn_process.state == "running":
            self.spawn_process.stop('stopped')
        self.dlg_benchmark.hide()

    def on_dlg_results_close(self, _widget, _event=None):
        """Handle btn_close_results.clicked and dlg_results.delete_event."""
        self.dlg_results.hide()



# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Commands used for guiplex.py <=> uiconnection.py AMP-based communication.
"""
from twisted.protocols import amp


class ClientConnected(amp.Command):
    """Command for guiplex.py->GUI.client_connected
     to remotely call uiconnection.py->Daemon.client_connected.
     """
    arguments = [(b'handle', amp.Unicode())]
    response = []


class ClientDisconnected(amp.Command):
    """Command for guiplex.py->GUI.client_disconnected
     to remotely call uiconnection.py->Daemon.client_disconnected.
     """
    arguments = [(b'handle', amp.Unicode())]
    response = []


class EnumerateClients(amp.Command):
    """Command for uiconnection.py->Daemon.enumerate_clients
     to remotely call guiplex.py->GUI.enumerate_clients.
     """
    arguments = []
    response = [(b'handles', amp.ListOf(amp.Unicode()))]


class ClientCommand(amp.Command):
    """Command for uiconnection.py->Daemon.command
     to remotely call guiplex.py->GUI.command.
     """
    arguments = [(b'handle', amp.Unicode()),
                 (b'command', amp.Unicode())]
    response = [(b'result', amp.String()),
                (b'filename', amp.Unicode())]



# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Classes for creating client and group objects.
"""
clients = []


class Client:
    """An epoptes client."""
    def __init__(self, type_='', mac='', hostname='', alias='', users=None,
                 hsystem=''):
        self.type = type_
        self.mac = mac.lower()
        self.hostname = hostname
        self.alias = alias
        if users is None:
            self.users = {}
        else:
            self.users = users.copy()
        self.hsystem = hsystem
        clients.append(self)

    def set_offline(self):
        """Indicate that a client is now offline."""
        self.hsystem = ''
        self.users = {}
        self.type = 'offline'
        self.hostname = ''

    def is_offline(self):
        """Return True only if there is no system or session handle for this
        client but there is a MAC address.
        """
        return not(self.users or self.hsystem) and self.mac

    def get_name(self):
        """Return the alias of the client or the hostname if the alias is not
        set, or if both are unset (offline clients) return the MAC address.
        """
        return self.alias or self.hostname or self.mac

    def set_name(self, name):
        """Set the client alias."""
        self.alias = name

    def add_user(self, username, realname, handle):
        """Add a user in the users dict."""
        self.users[handle] = {'uname': username, 'rname': realname}


class Group:
    """A group of epoptes clients."""
    def __init__(self, name, members):
        self.name = name
        # {<instance> : {'x_pos':342, 'y_pos':112, 'size':600}}
        self.members = members.copy()

    def get_members(self):
        """Return a list with all the clients that are members of the group."""
        return self.members.keys()

    def has_client(self, client):
        """Check if a client is member of the group."""
        return client in self.members

    def add_client(self, client, **props):
        """Add a client to the group."""
        self.members[client] = {}
        self.set_properties(client, **props)

    def remove_client(self, client):
        """Remove a client from the group."""
        del self.members[client]

    def set_properties(self, client, **props):
        """Set x_pos, y_pos or size property for a client on the group."""
        for prop, value in props.items():
            if isinstance(value, (int, float)):
                self.members[client][prop] = value
                # TODO: Save the new values to the disk.
            else:
                raise TypeError("The value of '%s' must be a number." % prop)

    def get_property(self, client, prop):
        """Return the value of the 'prop' property for client."""
        return self.members[client][prop]



#!/usr/bin/python3
# This file is part of Epoptes, https://epoptes.org
# Copyright 2010-2020 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Epoptes GUI class.
"""
from distutils.version import LooseVersion
import getpass
import locale
import os
import pipes
import random
import socket
import string
import subprocess

from epoptes.ui.reactor import reactor
from epoptes.common.constants import *
from epoptes.common import config
from epoptes.core import logger, structs, wol
from epoptes.ui.about import About
from epoptes.ui.benchmark import Benchmark
from epoptes.ui.client_information import ClientInformation
from epoptes.ui.common import gettext as _
from epoptes.ui.exec_command import ExecCommand
from epoptes.ui.notifications import NotifyQueue
from epoptes.ui.send_message import SendMessage

from gi.repository import Gdk, GdkPixbuf, GLib, Gtk

LOG = logger.Logger(__file__)


class EpoptesGui(object):
    """Epoptes GUI class."""
    def __init__(self):
        # Initialization of general-purpose variables
        self.about = None
        self.benchmark = None
        self.client_information = None
        self.current_macs = subprocess.Popen(
            ['sh', '-c',
             r"""ip -oneline -family inet link show | """
             r"""sed -n 's/.*ether[[:space:]]*\([[:xdigit:]:]*\).*/\1/p';"""
             r"""echo $LTSP_CLIENT_MAC"""],
            stdout=subprocess.PIPE).communicate()[0].decode().lower().split()
        self.current_thumbshots = dict()
        self.daemon = None
        self.displayed_compatibility_warning = False
        self.exec_command = None
        self.imagetypes = {
            'thin': GdkPixbuf.Pixbuf.new_from_file('images/thin.svg'),
            'fat': GdkPixbuf.Pixbuf.new_from_file('images/fat.svg'),
            'standalone': GdkPixbuf.Pixbuf.new_from_file(
                'images/standalone.svg'),
            'offline': GdkPixbuf.Pixbuf.new_from_file('images/offline.svg')}
        self.labels_order = (1, 0)
        self.notify_queue = NotifyQueue(
            'Epoptes',
            '/usr/share/icons/hicolor/scalable/apps/epoptes.svg')
        self.send_message = None
        self.show_real_names = config.settings.getboolean(
            'GUI', 'show_real_names', fallback=False)
        # Thumbshot width and height. Good width defaults are multiples of 16,
        # so that the height is an integer in both 16/9 and 4/3 aspect ratios.
        self.ts_width = config.settings.getint(
            'GUI', 'thumbshots_width', fallback=128)
        self.ts_height = int(self.ts_width/(4/3))
        self.uid = os.getuid()
        self.vncserver = None
        self.vncserver_port = None
        self.vncserver_pwd = None
        self.vncviewer = None
        self.vncviewer_port = None

        self.builder = Gtk.Builder()
        self.builder.add_from_file('epoptes.ui')
        self.get = self.builder.get_object
        # Hide the remote assistance menuitem if epoptes-client isn't installed
        if not os.path.isfile('/usr/share/epoptes-client/remote_assistance.py'):
            self.get('imi_help_remote_support').set_property('visible', False)
            self.get('smi_help_remote_support').set_property('visible', False)
        self.mnu_add_to_group = self.get('mnu_add_to_group')
        self.mni_add_to_group = self.get('mni_add_to_group')
        self.gstore = Gtk.ListStore(str, object, bool)
        self.trv_groups = self.get('trv_groups')
        self.trv_groups.set_model(self.gstore)
        self.mainwin = self.get('wnd_main')
        self.cstore = Gtk.ListStore(str, GdkPixbuf.Pixbuf, object, str)
        self.icv_clients = self.get('icv_clients')
        self.set_labels_order(1, 0, None)
        self.icv_clients.set_model(self.cstore)
        self.icv_clients.set_pixbuf_column(C_PIXBUF)
        self.icv_clients.set_text_column(C_LABEL)
        self.cstore.set_sort_column_id(C_LABEL, Gtk.SortType.ASCENDING)
        self.on_icv_clients_selection_changed(None)
        self.icv_clients.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            [Gtk.TargetEntry.new("add", Gtk.TargetFlags.SAME_APP, 0)],
            Gdk.DragAction.COPY)
        self.trv_groups.enable_model_drag_dest(
            [("add", Gtk.TargetFlags.SAME_APP, 0)], Gdk.DragAction.COPY)
        self.default_group = structs.Group(
            '<b>'+_('Detected clients')+'</b>', {})
        default_iter = self.gstore.append(
            [self.default_group.name, self.default_group, False])
        self.default_group_ref = Gtk.TreeRowReference.new(
            self.gstore, self.gstore.get_path(default_iter))
        # Connect glade handlers with the callback functions
        self.builder.connect_signals(self)
        self.trv_groups.get_selection().select_path(
            self.default_group_ref.get_path())
        self.get('adj_icon_size').set_value(self.ts_width)
        self.on_scl_icon_size_value_changed(None)
        # Support a global groups.json, writable only by the "administrator"
        self.groups_file = '/etc/epoptes/groups.json'
        if os.access(self.groups_file, os.R_OK):
            # Don't use global groups for the "administrator"
            self.global_groups = not os.access(self.groups_file, os.W_OK)
        else:
            self.groups_file = config.expand_filename('groups.json')
            self.global_groups = False
        try:
            _saved_clients, groups = config.read_groups(self.groups_file)
        except ValueError as exc:
            self.warning_dialog(
                _('Failed to read the groups file:') + '\n'
                + self.groups_file + '\n'
                + _('You may need to restore your groups from a backup!')
                + '\n\n' + str(exc))
            _saved_clients, groups = [], []
        # In global groups mode, groups that start with X- are hidden
        self.x_groups = {}
        if self.global_groups:
            for group in list(groups):
                if group.name.upper().startswith('X-HIDDEN'):
                    if 'X-HIDDEN' not in self.x_groups:
                        self.x_groups['X-HIDDEN'] = []
                    self.x_groups['X-HIDDEN'] = list(set(
                        self.x_groups['X-HIDDEN']
                        + [m.mac for m in group.members]))
                if group.name.upper().startswith('X-'):
                    groups.remove(group)
        if groups:
            self.mni_add_to_group.set_sensitive(True)
        for group in groups:
            self.gstore.append([group.name, group, True])
            mitem = Gtk.MenuItem(label=group.name)
            mitem.show()
            mitem.connect(
                'activate', self.on_imi_clients_add_to_group_activate, group)
            self.mnu_add_to_group.append(mitem)
        self.fill_icon_view(self.get_selected_group()[1])
        self.trv_groups.get_selection().select_path(
            config.settings.getint('GUI', 'selected_group', fallback=0))
        mitem = self.get(config.settings.get(
            'GUI', 'label', fallback='rmi_labels_host_user'))
        if not mitem:
            mitem = self.get('rmi_labels_host_user')
        mitem.set_active(True)
        self.get('cmi_show_real_names').set_active(self.show_real_names)
        self.mainwin.set_sensitive(False)

    def save_settings(self):
        """Helper function for on_imi_file_quit_activate."""
        sel_group = self.gstore.get_path(self.get_selected_group()[0])[0]
        self.gstore.remove(self.gstore.get_iter(
            self.default_group_ref.get_path()))
        if not self.global_groups:
            config.save_groups(self.groups_file, self.gstore)
        settings = config.settings
        if not settings.has_section('GUI'):
            settings.add_section('GUI')

        settings.set('GUI', 'selected_group', str(sel_group))
        settings.set('GUI', 'show_real_names', str(self.show_real_names))
        settings.set('GUI', 'thumbshots_width', str(self.ts_width))
        config.write_ini_file(config.expand_filename('settings'), settings)

    def on_imi_file_quit_activate(self, _widget):
        """Handle imi_file_quit.activate and wnd_main.destroy events."""
        self.save_settings()
        if self.vncserver is not None:
            self.vncserver.kill()
        if self.vncviewer is not None:
            self.vncviewer.kill()
        # noinspection PyUnresolvedReferences
        reactor.stop()

    def on_rmi_labels_host_user_toggled(self, rmi):
        """Handle rmi_labels_host_user.toggled event."""
        self.set_labels_order(1, 0, rmi)

    def on_rmi_labels_host_toggled(self, rmi):
        """Handle rmi_labels_host.toggled event."""
        self.set_labels_order(-1, 0, rmi)

    def on_rmi_labels_user_host_toggled(self, rmi):
        """Handle rmi_labels_user_host.toggled event."""
        self.set_labels_order(0, 1, rmi)

    def on_rmi_labels_user_toggled(self, rmi):
        """Handle rmi_labels_user.toggled event."""
        self.set_labels_order(0, -1, rmi)

    def set_labels_order(self, user_pos, name_pos, rmi):
        """Helper function for on_rmi_labels_*_toggled."""
        # Save the order so all new clients get the selected format
        if rmi:
            config.settings.set('GUI', 'label', Gtk.Buildable.get_name(rmi))
        self.labels_order = (user_pos, name_pos)
        for row in self.cstore:
            self.set_label(row)

    def on_cmi_show_real_names_toggled(self, widget):
        """Handle cmi_show_real_names.toggled event."""
        self.show_real_names = widget.get_active()
        for row in self.cstore:
            self.set_label(row)

    def on_imi_session_boot_activate(self, _widget):
        """Handle imi_session_boot.activate event."""
        clients = self.get_selected_clients()
        if not clients:  # No client selected, send the command to all
            clients = self.cstore
        for client in clients:
            # Make sure that only offline computers will be sent to wol
            client = client[C_INSTANCE]
            if client.is_offline():
                wol.wake_on_lan(client.mac)

    def on_imi_session_logout_activate(self, _widget):
        """Handle imi_session_logout.activate event."""
        self.exec_on_selected_clients(
            ['logout'], mode=EM_SESSION,
            warn=_('Are you sure you want to log off all the users?'))

    def on_imi_session_reboot_activate(self, _widget):
        """Handle imi_session_reboot.activate event."""
        self.exec_on_selected_clients(
            ["reboot"],
            warn=_('Are you sure you want to reboot all the computers?'))

    def on_imi_session_shutdown_activate(self, _widget):
        """Handle imi_session_shutdown.activate event."""
        self.exec_on_selected_clients(
            ["shutdown"],
            warn=_('Are you sure you want to shutdown all the computers?'))

    @staticmethod
    def find_unused_port():
        """Find an unused port."""
        sck = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sck.bind(('', 0))
        sck.listen(1)
        port = sck.getsockname()[1]
        sck.close()
        return port

    def reverse_connection(self, cmd, *args):
        """Helper function for on_imi_broadcasts_*_activate."""
        # Open vncviewer in listen mode
        if self.vncviewer is None or self.vncviewer.poll() is not None:
            self.vncviewer_port = self.find_unused_port()
            # If the user installed ssvnc, prefer it over xvnc4viewer
            if os.path.isfile('/usr/bin/ssvncviewer'):
                self.vncviewer = subprocess.Popen(
                    ['ssvncviewer', '-multilisten',
                     str(self.vncviewer_port-5500)])
            elif os.path.isfile('/usr/bin/xtigervncviewer'):
                self.vncviewer = subprocess.Popen(
                    ['xtigervncviewer', '-listen', str(self.vncviewer_port)])
            elif os.path.isfile('/usr/bin/xvnc4viewer'):
                self.vncviewer = subprocess.Popen(
                    ['xvnc4viewer', '-listen', str(self.vncviewer_port)])
            # Support tigervnc on rpm distributions (LP: #1501747)
            elif os.path.isfile('/usr/share/locale/de/LC_MESSAGES/tigervnc.mo'):
                self.vncviewer = subprocess.Popen(
                    ['vncviewer', '-listen', str(self.vncviewer_port)])
            # The rest of the viewers, like tightvnc
            else:
                self.vncviewer = subprocess.Popen(
                    ['vncviewer', '-listen', str(self.vncviewer_port-5500)])

        # And, tell the clients to connect to the server
        self.exec_on_selected_clients([cmd, self.vncviewer_port] + list(args))

    def on_imi_broadcasts_monitor_user_activate(self, _widget):
        """Handle imi_sbroadcasts_monitor_user.activate event."""
        self.reverse_connection('get_monitored')

    def on_imi_broadcasts_assist_user_activate(self, _widget,
                                               _path=None, _view_column=None):
        """Handle imi_sbroadcasts_assist_user.activate event."""
        if config.settings.getboolean('GUI', 'grabkbdptr', fallback=False):
            self.reverse_connection('get_assisted', 'True')
        else:
            self.reverse_connection('get_assisted')

    def broadcast_screen(self, fullscreen=''):
        """Helper function for on_imi_broadcasts_broadcast_screen*_activate."""
        if self.vncserver is None:
            pwdfile = config.expand_filename('vncpasswd')
            pwd = ''.join(random.sample(
                string.ascii_letters + string.digits, 8))
            subprocess.call(['x11vnc', '-storepasswd', pwd, pwdfile])
            with open(pwdfile, 'rb') as file:
                pwd = file.read()
            self.vncserver_port = self.find_unused_port()
            self.vncserver_pwd = ''.join('\\%o' % c for c in pwd)
            self.vncserver = subprocess.Popen(
                ['x11vnc', '-noshm', '-nopw', '-quiet', '-viewonly', '-shared',
                 '-forever', '-nolookup', '-24to32', '-threads', '-rfbport',
                 str(self.vncserver_port), '-rfbauth', pwdfile])
        # Running `xdg-screensaver reset` as root doesn't reset the D.E.
        # screensaver, so send the reset command to both epoptes processes
        self.exec_on_selected_clients(
            ['reset_screensaver'], mode=EM_SYSTEM_AND_SESSION)
        self.exec_on_selected_clients(
            ["receive_broadcast", self.vncserver_port, self.vncserver_pwd,
             fullscreen], mode=EM_SYSTEM_OR_SESSION)

    def on_imi_broadcasts_broadcast_screen_fullscreen_activate(self, _widget):
        """Handle imi_broadcasts_broadcast_screen_fullscreen.activate event."""
        self.broadcast_screen('true')

    def on_imi_broadcasts_broadcast_screen_windowed_activate(self, _widget):
        """Handle imi_broadcasts_broadcast_screen_windowed.activate event."""
        self.broadcast_screen('')

    def on_imi_broadcasts_stop_broadcasts_activate(self, _widget):
        """Handle imi_broadcasts_stop_broadcasts.activate event."""
        self.exec_on_clients(['stop_receptions'], self.cstore,
                             mode=EM_SYSTEM_AND_SESSION)
        if self.vncserver is not None:
            self.vncserver.kill()
            self.vncserver = None

    # TODO: Should we allow for running arbitrary commands in clients?
    def on_imi_execute_execute_command_activate(self, _widget):
        """Handle imi_execute_execute_command.activate event."""
        if not self.exec_command:
            self.exec_command = ExecCommand(self.mainwin)
        cmd = self.exec_command.run()
        # If Cancel or Close were clicked
        if cmd == '':
            return
        if cmd.startswith("sudo "):
            e_m = EM_SYSTEM
            cmd = cmd[5:]
        else:
            e_m = EM_SESSION
        self.exec_on_selected_clients(['execute', cmd], mode=e_m)

    def on_imi_execute_send_message_activate(self, _widget):
        """Handle imi_execute_send_message.activate event."""
        if not self.send_message:
            self.send_message = SendMessage(self.mainwin)
        params = self.send_message.run()
        if params:
            self.exec_on_selected_clients(['message'] + list(params))

    def open_terminal(self, e_m):
        """Helper function for on_imi_open_terminal_*_activate."""
        clients = self.get_selected_clients()
        # If there is no client selected, send the command to all
        if not clients:
            clients = self.cstore
        for client in clients:
            inst = client[C_INSTANCE]
            if inst.type == 'offline':
                continue
            port = self.find_unused_port()
            user = '--'
            if e_m == EM_SESSION and client[C_SESSION_HANDLE]:
                user = inst.users[client[C_SESSION_HANDLE]]['uname']
            elif e_m == EM_SYSTEM:
                user = 'root'
            title = '%s@%s' % (user, inst.get_name())
            subprocess.Popen(['xterm', '-T', title, '-e', 'socat',
                              'tcp-listen:%d,keepalive=1' % port,
                              'stdio,raw,echo=0'])
            self.exec_on_clients(['remote_term', port], [client], mode=e_m)

    def on_imi_open_terminal_user_locally_activate(self, _widget):
        """Handle imi_open_terminal_user_locally.activate event."""
        self.open_terminal(EM_SESSION)

    def on_imi_open_terminal_root_locally_activate(self, _widget):
        """Handle imi_open_terminal_root_locally.activate event."""
        self.open_terminal(EM_SYSTEM)

    def on_imi_open_terminal_root_remotely_activate(self, _widget):
        """Handle imi_open_terminal_root_remotely.activate event."""
        self.exec_on_selected_clients(['root_term'], mode=EM_SYSTEM)

    def on_imi_restrictions_lock_screen_activate(self, _widget):
        """Handle imi_restrictions_lock_screen.activate event."""
        msg = _("The screen is locked by a system administrator.")
        self.exec_on_selected_clients(['lock_screen', 0, msg])

    def on_imi_restrictions_unlock_screen_activate(self, _widget):
        """Handle imi_restrictions_unlock_screen.activate event."""
        self.exec_on_selected_clients(
            ['unlock_screen'], mode=EM_SESSION_AND_SYSTEM)

    def on_imi_restrictions_mute_sound_activate(self, _widget):
        """Handle imi_restrictions_mute_sound.activate event."""
        self.exec_on_selected_clients(
            ['mute_sound', 0], mode=EM_SYSTEM_AND_SESSION)

    def on_imi_restrictions_unmute_sound_activate(self, _widget):
        """Handle imi_restrictions_unmute_sound.activate event."""
        self.exec_on_selected_clients(
            ['unmute_sound'], mode=EM_SYSTEM_AND_SESSION)

    def on_imi_clients_add_to_group_activate(self, _widget, group):
        """Handle *dynamic* imi_clients_add_to_group.activate event."""
        clients = self.get_selected_clients()
        for client in clients:
            if not group.has_client(client[C_INSTANCE]):
                group.add_client(client[C_INSTANCE])

    def on_imi_clients_remove_from_group_activate(self, _widget):
        """Handle imi_clients_remove_from_group.activate event."""
        clients = self.get_selected_clients()
        group = self.get_selected_group()[1]
        if self.confirmation_dialog(
                _('Are you sure you want to remove the selected client(s)'
                  ' from group "%s"?') % group.name):
            for client in clients:
                group.remove_client(client[C_INSTANCE])
            self.fill_icon_view(self.get_selected_group()[1], True)

    def on_imi_clients_network_benchmark_activate(self, _widget):
        """Handle imi_clients_network_benchmark.activate event."""
        if not self.benchmark:
            self.benchmark = Benchmark(self.mainwin, self.daemon.command)
        self.benchmark.run(self.get_selected_clients() or self.cstore)

    def on_imi_clients_information_activate(self, _widget):
        """Handle imi_clients_information.activate event."""
        if not self.client_information:
            self.client_information = ClientInformation(self.mainwin)
        self.client_information.btn_edit_alias.set_sensitive(
            not self.is_default_group_selected())
        self.client_information.run(
            self.get_selected_clients()[0], self.daemon.command)
        self.set_label(self.get_selected_clients()[0])

    @staticmethod
    def open_url(link):
        """Helper function for on_imi_open_terminal_*_activate."""
        subprocess.Popen(["xdg-open", link])

    def on_imi_help_home_activate(self, _widget):
        """Handle imi_help_home.activate event."""
        self.open_url("https://epoptes.org")

    def on_imi_help_report_bug_activate(self, _widget):
        """Handle imi_help_report_bug.activate event."""
        self.open_url("https://bugs.launchpad.net/epoptes")

    def on_imi_help_ask_question_activate(self, _widget):
        """Handle imi_help_ask_question.activate event."""
        self.open_url("https://answers.launchpad.net/epoptes")

    def on_imi_help_translate_application_activate(self, _widget):
        """Handle imi_help_translate_application.activate event."""
        self.open_url("https://epoptes.org/translations")

    def on_imi_help_live_chat_irc_activate(self, _widget):
        """Handle imi_help_live_chat_irc.activate event."""
        host = socket.gethostname()
        user = getpass.getuser()
        lang = locale.getlocale()[0]
        self.open_url("http://ts.sch.gr/repo/irc?user=%s&host=%s&lang=%s" %
                      (user, host, lang))

    @staticmethod
    def on_imi_help_remote_support_activate(_widget):
        """Handle imi_help_remote_support.activate event."""
        subprocess.Popen('./remote_assistance.py',
                         cwd='/usr/share/epoptes-client')

    def on_imi_help_about_activate(self, _widget):
        """Handle imi_help_about_activate.activate event."""
        if not self.about:
            self.about = About(self.mainwin)
        self.about.run()

    def on_trv_groups_drag_drop(self, _wid, context, drag_x, drag_y, time):
        """Handle trv_groups.drag_drop event."""
        dest = self.trv_groups.get_dest_row_at_pos(drag_x, drag_y)
        if dest is not None:
            path, _pos = dest
            group = self.gstore[path][G_INSTANCE]
            if group is not self.default_group:
                for cln in self.get_selected_clients():
                    cln = cln[C_INSTANCE]
                    if not group.has_client(cln):
                        group.add_client(cln)

        context.finish(True, False, time)
        return True

    def on_trv_groups_drag_motion(self, widget, context, drag_x, drag_y, time):
        """Handle trv_groups.drag_motion event."""
        # Don't allow dropping in the empty space of the treeview, or inside
        # the 'Detected clients' group, or inside the currently selected group
        drag_info = widget.get_dest_row_at_pos(drag_x, drag_y)
        act = 0
        if drag_info:
            path, pos = drag_info
            if pos:
                if path != self.gstore.get_path(self.get_selected_group()[0]) \
                        and path != self.default_group_ref.get_path():
                    act = context.get_suggested_action()
        Gdk.drag_status(context, act, time)
        return True

    def on_crt_group_edited(self, _widget, path, new_name):
        """Handle crt_group.edited event."""
        self.gstore[path][G_LABEL] = new_name
        self.gstore[path][G_INSTANCE].name = new_name
        self.mnu_add_to_group.get_children()[int(path)-1].set_label(new_name)

    def on_trs_groups_changed(self, _treeselection):
        """Handle trs_groups.changed event."""
        self.cstore.clear()
        selected = self.get_selected_group()

        if selected is not None:
            self.fill_icon_view(selected[1])
            path = self.gstore.get_path(selected[0])[0]
            self.mnu_add_to_group.foreach(lambda w: w.set_sensitive(True))
            menuitems = self.mnu_add_to_group.get_children()
            if path != 0 and path-1 < len(menuitems):
                menuitems[path-1].set_sensitive(False)
        else:
            if not self.default_group_ref.valid():
                return
            self.trv_groups.get_selection().select_path(
                self.default_group_ref.get_path())
        self.get('btn_group_remove').set_sensitive(
            not self.is_default_group_selected())
        self.set_move_group_sensitivity()

    def set_move_group_sensitivity(self):
        """Helper function for on_btn_group_*_clicked."""
        selected = self.get_selected_group()
        selected_path = self.gstore.get_path(selected[0])[0]
        blocker = not selected[1] is self.default_group
        self.get('btn_group_up').set_sensitive(blocker and selected_path > 1)
        self.get('btn_group_down').set_sensitive(
            blocker and selected_path < len(self.gstore)-1)

    def on_btn_group_add_clicked(self, _widget):
        """Handle btn_group_add.clicked event."""
        new_group = structs.Group(_('New group'), {})
        itr = self.gstore.append([new_group.name, new_group, True])
        # Edit the name of the newly created group
        self.trv_groups.set_cursor(
            self.gstore.get_path(itr), self.get('tvc_group'), True)
        menuitem = Gtk.MenuItem(new_group.name)
        menuitem.show()
        menuitem.connect(
            'activate', self.on_imi_clients_add_to_group_activate, new_group)
        self.mnu_add_to_group.append(menuitem)

    def on_btn_group_remove_clicked(self, _widget):
        """Handle btn_group_remove.clicked event."""
        group_iter = self.get_selected_group()[0]
        group = self.gstore[group_iter][G_INSTANCE]

        if self.confirmation_dialog(
                _('Are you sure you want to remove group "%s"?') % group.name):
            path = self.gstore.get_path(group_iter)[0]
            self.gstore.remove(group_iter)
            menuitem = self.mnu_add_to_group.get_children()[path-1]
            self.mnu_add_to_group.remove(menuitem)

    def on_btn_group_up_clicked(self, _widget):
        """Handle btn_group_up.clicked event."""
        selected_group_iter = self.get_selected_group()[0]
        path = self.gstore.get_path(selected_group_iter)[0]
        previous_iter = self.gstore.get_iter(path-1)
        self.gstore.swap(selected_group_iter, previous_iter)
        self.set_move_group_sensitivity()
        mitem = self.mnu_add_to_group.get_children()[path-1]
        self.mnu_add_to_group.reorder_child(mitem, path-2)

    def on_btn_group_down_clicked(self, _widget):
        """Handle btn_group_down.clicked event."""
        selected_group_iter = self.get_selected_group()[0]
        path = self.gstore.get_path(selected_group_iter)[0]
        self.gstore.swap(selected_group_iter,
                         self.gstore.iter_next(selected_group_iter))
        self.set_move_group_sensitivity()
        mitem = self.mnu_add_to_group.get_children()[path-1]
        self.mnu_add_to_group.reorder_child(mitem, path)

    def on_icv_clients_selection_changed(self, _widget):
        """Handle icv_clients.selection_changed event."""
        selected = self.get_selected_clients()
        single_client = False
        if len(selected) == 1:
            single_client = True
        self.get('imi_clients_information').set_sensitive(single_client)
        self.get('tlb_clients_information').set_sensitive(single_client)

        if selected:
            self.get('mni_add_to_group').set_sensitive(True)
            self.get('imi_clients_remove_from_group').set_sensitive(
                not self.is_default_group_selected())
        else:
            self.get('mni_add_to_group').set_sensitive(False)
            self.get('imi_clients_remove_from_group').set_sensitive(False)

        if len(selected) > 1:
            self.get('lbl_status').set_text(
                _('%d clients selected' % len(selected)))
        else:
            self.get('lbl_status').set_text('')

    def on_icv_clients_button_press_event(self, widget, event):
        """Handle icv_clients.button_press event."""
        clicked = widget.get_path_at_pos(int(event.x), int(event.y))

        if event.button == 3:
            if widget is self.icv_clients:
                selection = widget
                selected = widget.get_selected_items()
            else:
                selection = widget.get_selection()
                selected = selection.get_selected_rows()[1]
                if clicked:
                    clicked = clicked[0]
            if clicked:
                if clicked not in selected:
                    selection.unselect_all()
                    selection.select_path(clicked)
            else:
                selection.unselect_all()
            if widget is self.icv_clients:
                menu = self.get('mni_clients').get_submenu()
                menu.popup(None, None, None, None, event.button, event.time)
                menu.show()
            return True
        return False

    def on_btn_size_down_clicked(self, _widget):
        """Handle btn_size_down.clicked event."""
        adj = self.get('adj_icon_size')
        adj.set_value(adj.get_value() - adj.props.page_increment)

    def on_btn_size_up_clicked(self, _widget):
        """Handle btn_size_up.clicked event."""
        adj = self.get('adj_icon_size')
        adj.set_value(adj.get_value() + adj.props.page_increment)

    def on_scl_icon_size_value_changed(self, _widget):
        """Handle scl_icon_size.value_changed event."""
        self.ts_width = int(self.get('adj_icon_size').get_value())
        self.ts_height = int(self.ts_width/(4/3))  # Κeep the 4:3 aspect ratio

        # Fast scale all the thumbshots to make the change quickly visible
        old_pixbufs = list(self.imagetypes.values())
        self.imagetypes = {
            'offline': GdkPixbuf.Pixbuf.new_from_file_at_size(
                'images/offline.svg', self.ts_width, self.ts_height),
            'thin': GdkPixbuf.Pixbuf.new_from_file_at_size(
                'images/thin.svg', self.ts_width, self.ts_height),
            'fat': GdkPixbuf.Pixbuf.new_from_file_at_size(
                'images/fat.svg', self.ts_width, self.ts_height),
            'standalone': GdkPixbuf.Pixbuf.new_from_file_at_size(
                'images/standalone.svg', self.ts_width, self.ts_height),
        }
        for row in self.cstore:
            if row[C_PIXBUF] in old_pixbufs:
                row[C_PIXBUF] = self.imagetypes[row[C_INSTANCE].type]
            else:
                new_thumb = row[C_PIXBUF].scale_simple(
                    self.ts_width, self.ts_height,
                    GdkPixbuf.InterpType.NEAREST)
                row[C_PIXBUF] = new_thumb

        # Hack to remove the extra padding that remains after a 'zoom out'
        # TODO: Weird Gtk3 issue, they calculate excess width:
        # https://bugzilla.gnome.org/show_bug.cgi?id=680953
        # https://stackoverflow.com/questions/14090094/what-causes-the-different-display-behaviour-for-a-gtkiconview-between-different
        self.icv_clients.get_cells()[0].set_fixed_size(self.ts_width/4, -1)
        self.icv_clients.check_resize()

    def on_scl_icon_size_button_press_event(self, _widget, event):
        """Make right click reset the thumbshots size."""
        if event.button == 3:
            self.get('adj_icon_size').set_value(128)
            return True
        return False

    # Daemon callbacks
    def connected(self, daemon):
        """Called from uiconnection->Daemon->connectionMade."""
        self.mainwin.set_sensitive(True)
        self.daemon = daemon
        daemon.enumerate_clients().addCallback(self.amp_got_clients)
        self.fill_icon_view(self.get_selected_group()[1])

    def disconnected(self, _daemon):
        """Called from uiconnection->Daemon->connectionLost."""
        self.mainwin.set_sensitive(False)
        # If the reactor is not running at this point it means that we were
        # closed normally.
        # noinspection PyUnresolvedReferences
        if not reactor.running:
            return
        self.save_settings()
        msg = _("Lost connection with the epoptes service.")
        msg += "\n\n" + \
               _("Make sure the service is running and then restart epoptes.")
        dlg = Gtk.MessageDialog(type=Gtk.MessageType.ERROR,
                                buttons=Gtk.ButtonsType.OK, message_format=msg)
        dlg.set_title(_('Service connection error'))
        dlg.run()
        dlg.destroy()
        # noinspection PyUnresolvedReferences
        reactor.stop()

    def amp_client_connected(self, handle):
        """Called from uiconnection->Daemon->client_connected."""
        LOG.w("New connection from", handle)
        dfr = self.daemon.command(handle, 'info')
        dfr.addCallback(lambda r: self.add_client(handle, r.decode()))
        dfr.addErrback(lambda err: LOG.e(
            "Error when connecting client %s: %s" % (handle, err)))

    def amp_client_disconnected(self, handle):
        """Called from uiconnection->Daemon->client_disconnected."""
        def determine_offline(client_):
            """Helper function to call client.set_offline when appropriate."""
            if client_.hsystem == '' and client_.users == {}:
                client_.set_offline()

        LOG.w("Disconnect from", handle)
        client = None
        for client in structs.clients:
            if client.hsystem == handle:
                if self.get_selected_group()[1].has_client(client) \
                        or self.is_default_group_selected():
                    self.notify_queue.enqueue(
                        _("Shut down:"), client.get_name())
                client.hsystem = ''
                determine_offline(client)
                break
            elif handle in client.users:
                if self.get_selected_group()[1].has_client(client) \
                        or self.is_default_group_selected():
                    self.notify_queue.enqueue(
                        _("Disconnected:"),
                        _("%(user)s from %(host)s") %
                        {"user": client.users[handle]['uname'],
                         "host": client.get_name()})
                del client.users[handle]
                determine_offline(client)
                break
            else:
                client = None
        if client is not None:
            for row in self.cstore:
                if row[C_INSTANCE] is client:
                    self.fill_icon_view(self.get_selected_group()[1], True)
                    break

    def amp_got_clients(self, handles):
        """Callback from self.connected=>daemon.enumerate_clients."""
        LOG.w("Got clients:", ', '.join(handles) or 'None')
        for handle in handles:
            dfr = self.daemon.command(handle, 'info')
            dfr.addCallback(
                lambda r, h=handle: self.add_client(h, r.decode(), True))
            dfr.addErrback(lambda err, h=handle: LOG.e(
                "Error when enumerating client %s: %s" % (h, err)))

    def add_to_icon_view(self, client):
        """Properly add a Client class instance to the clients iconview."""
        # If there are one or more users on client, add a new iconview entry
        # for each one of them.
        label = 'uname'
        if self.show_real_names:
            label = 'rname'
        if client.users:
            for hsession, user in client.users.items():
                self.cstore.append(
                    [self.calculate_label(client, user[label]),
                     self.imagetypes[client.type], client, hsession])
                self.ask_thumbshot(hsession, True)
        else:
            self.cstore.append(
                [self.calculate_label(client),
                 self.imagetypes[client.type], client, ''])

    def fill_icon_view(self, group, keep_selection=False):
        """Fill the clients iconview from a Group class instance."""
        if keep_selection:
            selection = [row[C_INSTANCE] for row in
                         self.get_selected_clients()]
        else:
            selection = None
        self.cstore.clear()
        if self.is_default_group_selected():
            clients_list = [client for client in structs.clients
                            if client.type != 'offline']
        else:
            clients_list = group.get_members()
        # Add the new clients to the iconview
        for client in clients_list:
            self.add_to_icon_view(client)
        if selection:
            for row in self.cstore:
                if row[C_INSTANCE] in selection:
                    self.icv_clients.select_path(row.path)
                    selection.remove(row[C_INSTANCE])

    def is_default_group_selected(self):
        """Return True if the default group is selected"""
        if self.get_selected_group():
            return self.get_selected_group()[1] is self.default_group
        return True

    def get_selected_group(self):
        """Return a 2-tuple containing the itr and the instance
        for the currently selected group."""
        itr = self.trv_groups.get_selection().get_selected()[1]
        if itr:
            return itr, self.gstore[itr][G_INSTANCE]
        return None

    def add_client(self, handle, reply, already=False):
        """Callback after running `info` on a client."""
        # already is True if the client was started before epoptes
        LOG.w("add_client's been called for", handle)
        try:
            info = {}
            for line in reply.strip().split('\n'):
                key, value = line.split('=', 1)
                info[key.strip()] = value.strip()
            user, host, _ip, mac, type_, uid, version, name = \
                info['user'], info['hostname'], info['ip'], info['mac'], \
                info['type'], int(info['uid']), info['version'], info['name']
        except (KeyError, ValueError) as exc:
            LOG.e("  Can't extract client information, won't add this client",
                  exc)
            return False

        # Support hiding clients in an X-Hidden group
        if self.global_groups and 'X-HIDDEN' in self.x_groups:
            if mac in self.x_groups['X-HIDDEN']:
                LOG.w("  X-HIDDEN: Won't add this client to my lists")
                return False

        # Check if the incoming client is the same with the computer in which
        # epoptes is running, so we don't add it to the list.
        if (mac in self.current_macs) and ((uid == self.uid) or (uid == 0)):
            LOG.w("  SAME-PC: Won't add this client to my lists")
            return False

        # Compatibility check
        if LooseVersion(version) < LooseVersion(COMPATIBILITY_VERSION):
            self.daemon.command(
                handle, "die 'Incompatible Epoptes server version!'")
            if not self.displayed_compatibility_warning:
                self.displayed_compatibility_warning = True
                self.warning_dialog(_(
                    """A connection attempt was made by a client with"""
                    """ version %s, which is incompatible with the current"""
                    """ epoptes version.\n\nYou need to update your clients"""
                    """ to the latest epoptes-client version.""") % version)
            return False
        sel_group = self.get_selected_group()[1]
        client = None
        for inst in structs.clients:
            # Find if the new handle is a known client
            if mac == inst.mac:
                client = inst
                LOG.w('  Old client: ', end='')
                break
        if client is None:
            LOG.w('  New client: ', end='')
            client = structs.Client(mac=mac)
        LOG.w('hostname=%s, type=%s, uid=%s, user=%s' %
              (host, type_, uid, user))

        # Update/fill the client information
        client.type, client.hostname = type_, host
        if uid == 0:
            # This is a root epoptes-client
            client.hsystem = handle
        else:
            # This is a user epoptes-client
            client.add_user(user, name, handle)
            if not already and (sel_group.has_client(client)
                                or self.is_default_group_selected()):
                self.notify_queue.enqueue(
                    _("Connected:"),
                    _("%(user)s on %(host)s") % {"user": user, "host": host})
        if sel_group.has_client(client) or self.is_default_group_selected():
            self.fill_icon_view(sel_group, True)

        return True

    def set_label(self, row):
        """Set the appropriate label for a client icon."""
        inst = row[C_INSTANCE]
        if row[C_SESSION_HANDLE]:
            label = 'uname'
            if self.show_real_names:
                label = 'rname'
            user = row[C_INSTANCE].users[row[C_SESSION_HANDLE]][label]
        else:
            user = ''
        row[C_LABEL] = self.calculate_label(inst, user)

    def calculate_label(self, client, username=''):
        """Return the iconview label from a hostname/alias and a username,
        according to the user options.
        """
        user_pos, name_pos = self.labels_order

        alias = client.get_name()
        if username == '' or user_pos == -1:
            return alias
        else:
            if user_pos == 0:
                label = username
                if name_pos == 1:
                    label += " (%s)" % alias
            else:  # name_pos == 0
                label = alias
                if user_pos == 1:
                    label += " (%s)" % username
            return label

    def ask_thumbshot(self, handle, first_time=False):
        """Ask a client for a thumbshot, every 5 seconds."""
        # Should always return False to prevent glib from calling us again
        if first_time:
            if handle not in self.current_thumbshots:
                # We started asking for thumbshots, but didn't yet get one
                self.current_thumbshots[handle] = None
            else:
                # Reuse the existing thumbshot
                if not self.current_thumbshots[handle] is None:
                    for i in self.cstore:
                        if handle == i[C_SESSION_HANDLE]:
                            self.cstore[i.path][C_PIXBUF] = \
                                self.current_thumbshots[handle]
                            break
                return False
        # TODO: Implement this using Gtk.TreeRowReferences instead of
        # searching the whole model (Need to modify exec_on_clients)
        for client in self.cstore:
            if handle == client[C_SESSION_HANDLE]:
                self.exec_on_clients(
                    ['thumbshot', self.ts_width, self.ts_height],
                    handles=[handle], reply=self.got_thumbshot)
                return False
        # That handle is no longer in the cstore, remove it
        if handle in self.current_thumbshots:
            del self.current_thumbshots[handle]
        return False

    def got_thumbshot(self, handle, reply):
        """Callback after running`thumbshot` on a client."""
        for i in self.cstore:
            if handle == i[C_SESSION_HANDLE]:
                # Ask for a new thumbshot THUMBSHOT_MS msec after the last one.
                # So if the client is too stressed and needs e.g. 7 secs to
                # send a thumbshot, we'll ask for one every 12 secs.
                GLib.timeout_add(int(config.system['THUMBSHOT_MS']),
                    self.ask_thumbshot, handle)
                LOG.d("I got a thumbshot from %s." % handle)
                if not reply:
                    return
                try:
                    rowstride, size, pixels = reply.split(b'\n', 2)
                    rowstride = int(rowstride)
                    width, height = [int(i) for i in size.split(b'x')]
                except ValueError:
                    LOG.e("Bad thumbshot header")
                    return
                # GLib.Bytes.new and .copy() avoid memory leak and crash (#110)
                pxb = GdkPixbuf.Pixbuf.new_from_bytes(
                    GLib.Bytes.new(pixels), GdkPixbuf.Colorspace.RGB, False, 8,
                    width, height, rowstride)
                pxb = pxb.copy()
                self.current_thumbshots[handle] = pxb
                self.cstore[i.path][C_PIXBUF] = pxb
                return
        # That handle is no longer in the cstore, remove it
        if handle in self.current_thumbshots:
            del self.current_thumbshots[handle]

    def get_selected_clients(self):
        """Return a list of the clients selected in the iconview."""
        selected = self.icv_clients.get_selected_items()
        items = []
        for i in selected:
            items.append(self.cstore[i])
        return items

    def exec_on_clients(
            self, command, clients=None, reply=None, mode=EM_SESSION_OR_SYSTEM,
            handles=None, warning='', params=None):
        """Execute a command on all (or on the provided) clients."""
        # reply should be a method in which the result will be sent
        if clients is None:
            clients = []
        if handles is None:
            handles = []
        if params is None:
            params = []
        # "if self.cstore" is wrong as it's a Gtk.ListStore, not a dict.
        # pylint: disable=len-as-condition
        if len(self.cstore) == 0:
            LOG.d('No clients')
            return

        if isinstance(command, list) and len(command) > 0:
            command = '%s %s' % (command[0], ' '.join(
                [pipes.quote(str(x)) for x in command[1:]]))

        if (clients != [] or handles != []) and warning != '':
            if not self.confirmation_dialog(warning):
                return
        if clients == [] and handles != []:
            for handle in handles:
                cmd = self.daemon.command(handle, str(command))
                # TODO: do we need errbacks even when no reply?
                if reply:
                    cmd.addCallback(
                        lambda r, h=handle, p=params: reply(h, r, *p))
                    cmd.addErrback(lambda err, h=handle: LOG.e(
                        "Error when executing command %s on client %s: %s" %
                        (command, h, err)))

        for client in clients:
            sent = False
            for e_m in mode:
                if e_m == EM_SESSION_ONLY:
                    handle = client[C_SESSION_HANDLE]
                elif e_m == EM_SYSTEM_ONLY:
                    handle = client[C_INSTANCE].hsystem
                else:  # e_m == EM_EXIT_IF_SENT
                    if sent:
                        break
                    else:
                        continue
                if handle == '':
                    continue
                sent = True
                cmd = self.daemon.command(handle, str(command))
                # TODO: do we need errbacks even when no reply?
                if reply:
                    cmd.addCallback(
                        lambda r, h=handle, p=params: reply(h, r, *p))
                    cmd.addErrback(lambda err: LOG.e(
                        "Error when executing command %s on client %s: %s" %
                        (command, handle, err)))

    def exec_on_selected_clients(
            self, command, reply=None, mode=EM_SESSION_OR_SYSTEM, warn=''):
        """Execute a command on the selected clients."""
        clients = self.get_selected_clients()
        if not clients:  # No client selected, send the command to all
            clients = self.cstore
        else:  # Show the warning only when no clients are selected
            warn = ''
        self.exec_on_clients(command, clients, reply, mode, warning=warn)

    def confirmation_dialog(self, text):
        """Show a Yes/No dialog, returning True/False accordingly."""
        dlg = Gtk.MessageDialog(
            self.mainwin, 0, Gtk.MessageType.WARNING, Gtk.ButtonsType.YES_NO,
            text, title=_('Confirm action'))
        resp = dlg.run()
        dlg.destroy()
        return resp == Gtk.ResponseType.YES

    def warning_dialog(self, text):
        """Show a warning dialog."""
        dlg = Gtk.MessageDialog(
            self.mainwin, 0, Gtk.MessageType.WARNING, Gtk.ButtonsType.CLOSE,
            text, title=_('Warning'))
        dlg.run()
        dlg.destroy()



#!/usr/bin/python3
# This file is part of Epoptes, https://epoptes.org
# Copyright 2018 the Epoptes team, see AUTHORS.
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Spawn a process asynchronously and get its output in a callback.
https://twistedmatrix.com/documents/current/core/howto/process.html
The spawned process may be terminated:
 * By itself with exit code = 0 (success)
 * By itself with exit code != 0 (failure)
 * By SpawnProcess when lines_max is reached (success)
 * By SpawnProcess when timeout is reached (failure)
 * By the caller (e.g. a cancel button) using SpawnProcess.stop() (failure)
It's possible that some useful stdout exists even on failure.
"""
from twisted.internet import protocol, reactor

from epoptes.core import logger


LOG = logger.Logger(__file__)


class SpawnProcess(protocol.ProcessProtocol):
    """Spawn a process asynchronously and get its output in a callback."""
    def __init__(self, on_exit):
        self.dc_stop = None  # IDelayedCall, to cancel the stop callback
        self.dc_timeout = None  # IDelayedCall, to cancel the timeout callback
        self.err_data = b''
        self.lines_count = 0
        self.lines_max = 0
        self.on_exit = on_exit
        self.out_data = b''
        self.reason = ''
        self.state = 'idle'  # or: running, stopping1, stopping2, stopping3

    def __del__(self):
        """Kill child processes even on abrupt exits."""
        LOG.d("__del__:")
        if self.state == "running":
            self.stop('destructor')

    def spawn(self, cmdline, timeout=0, lines_max=0):
        """Spawn the process and run on_exit on exit, timeout, or lines_max."""
        LOG.d("spawn: {}, timeout: {}, lines_max: {}".format(
            cmdline, timeout, lines_max))
        assert self.state == "idle"
        # Reinitialize vars without duplicating a lot of code
        self.__init__(self.on_exit)
        self.lines_max = lines_max
        if timeout:
            # noinspection PyUnresolvedReferences
            self.dc_timeout = reactor.callLater(timeout, self.stop, "timeout")
        else:
            self.dc_timeout = None
        # noinspection PyUnresolvedReferences
        reactor.spawnProcess(self, cmdline[0], cmdline)

    def stop(self, reason=""):
        """Send TERM, TERM, KILL, with 0.5 sec delays."""
        LOG.d("stop('{}')".format(reason))
        assert self.state != "idle"
        assert self.transport
        # Once stopping started, only accept calls from stop itself,
        # and not for example lines_max calls.
        if self.state.startswith("stopping") \
                and not reason.startswith("stopping"):
            return
        # Only keep the initial reason
        if not self.reason:
            self.reason = reason
        # Cancel the timeout callback
        if self.dc_timeout:
            if self.reason != "timeout":
                self.dc_timeout.cancel()
            self.dc_timeout = None
        if self.state == "running":
            self.state = "stopping1"
            self.transport.signalProcess('TERM')
        elif self.state == "stopping1":
            self.state = "stopping2"
            # iperf2 requests 2 TERM signals to abort waiting open connections
            # Additionally, help it by closing its stdio
            self.transport.loseConnection()
            self.transport.signalProcess('TERM')
        elif self.state == "stopping2":
            self.state = "stopping3"
            # Kill it, don't wait any longer
            self.transport.signalProcess('KILL')
        else:
            assert False, "Unable to kill child process!"
        # noinspection PyUnresolvedReferences
        self.dc_stop = reactor.callLater(0.5, self.stop, self.state)

    def connectionMade(self):
        """Override BaseProtocol.connectionMade."""
        LOG.d("connectionMade: pid =", self.transport.pid)
        self.state = 'running'

    def errReceived(self, data):
        """Override ProcessProtocol.errReceived."""
        LOG.d("errReceived:", data)
        self.err_data += data

    def outReceived(self, data):
        """Override ProcessProtocol.outReceived."""
        old_lines_count = self.lines_count
        self.lines_count += data.count(b'\n')
        LOG.d("outReceived, lines_count = %s:" % self.lines_count, data)
        self.out_data += data
        # If this is the first time lines_max is exceeded, call stop
        if self.lines_count >= self.lines_max > old_lines_count:
            self.stop("lines_max")

    def processExited(self, reason):
        """Override ProcessProtocol.processExited."""
        LOG.d("processExited:", reason.value.exitCode)
        self.state = "idle"
        # Cancel the timeout and stop callbacks
        if self.dc_timeout:
            self.dc_timeout.cancel()
            self.dc_timeout = None
        if self.dc_stop:
            self.dc_stop.cancel()
            self.dc_stop = None
        if not self.reason:
            self.reason = str(reason.value)
        # noinspection PyUnresolvedReferences
        reactor.callLater(0, self.call_on_exit)

    def call_on_exit(self):
        """Call on_exit after processing all twisted events.
        When e.g. the iperf port is already in use, processExited happens
        before errReceived gets "bind failed: Address already in use".
        Using callLater, on_exit is able to properly get and display stderr.
        """
        self.on_exit(self.out_data, self.err_data, self.reason)


def main():
    """Run a test from the command line."""

    def on_test_exit(out_data, err_data, reason):
        """Called by SpawnProcess when the test process exits."""
        LOG.d("  Callback out_data:", out_data)
        LOG.d("  Callback err_data:", err_data)
        LOG.d("  Callback reason:", reason)
        # Quit a bit later in order to test if events occur properly
        # noinspection PyUnresolvedReferences
        reactor.callLater(3, reactor.stop)

    script = """# Produce some output with delays
i=0
while [ $i -lt 5 ]; do
    echo "process stdout line #$i"
    echo "process stderr line #$i" >&2
    sleep 1
    i=$(($i+1))
done
"""
    tests = ['process-error']
    print("Running tests:", tests)
    if 'normal-termination' in tests:
        SpawnProcess(on_test_exit).spawn(['sh', '-c', script])
    if 'lines_max' in tests:
        SpawnProcess(on_test_exit).spawn(
            ['sh', '-c', 'trap "" TERM\n' + script], lines_max=3)
    if 'timeout' in tests:
        SpawnProcess(on_test_exit).spawn(['sh', '-c', script], timeout=3)
    if 'event-driven' in tests:  # e.g. a cancel button
        proc = SpawnProcess(on_test_exit)
        proc.spawn(['sh', '-c', script])
        # noinspection PyUnresolvedReferences
        reactor.callLater(3, proc.stop, 'cancelled')
    if 'process-error' in tests:
        SpawnProcess(on_test_exit).spawn(
            ['sh', '-c', 'echo Started && sleep 1 && invalid-command'])

    # noinspection PyUnresolvedReferences
    reactor.run()


if __name__ == '__main__':
    main()
